@@ -1,4 +1,5 @@
-require 'win32/importer'
+# frozen_string_literal: true
+require 'fiddle/import'
 
 module Win32
 
@@ -68,7 +69,11 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
   WCHAR_NUL = "\0".encode(WCHAR).freeze
   WCHAR_CR = "\r".encode(WCHAR).freeze
   WCHAR_SIZE = WCHAR_NUL.bytesize
-  LOCALE = Encoding.find(Encoding.locale_charmap)
+  begin
+    LOCALE = Encoding.find(Encoding.locale_charmap)
+  rescue ArgumentError
+    LOCALE = Encoding::UTF_8
+  end
 
   class Registry
 
@@ -168,7 +173,7 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
     #
     class Error < ::StandardError
       module Kernel32
-        extend Importer
+        extend Fiddle::Importer
         dlload "kernel32.dll"
       end
       FormatMessageW = Kernel32.extern "int FormatMessageW(int, void *, int, int, void *, int, void *)", :stdcall
@@ -224,7 +229,7 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
     #
     module API
       include Constants
-      extend Importer
+      extend Fiddle::Importer
       dlload "advapi32.dll"
       [
         "long RegOpenKeyExW(void *, void *, long, long, void *)",
@@ -253,30 +258,34 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
         /^(?:x64|x86_64)/ =~ RUBY_PLATFORM
       end
 
+      TEMPLATE_HANDLE = 'J<'
+
       def packhandle(h)
-        win64? ? packqw(h) : packdw(h)
+        [h].pack(TEMPLATE_HANDLE)
       end
 
       def unpackhandle(h)
-        win64? ? unpackqw(h) : unpackdw(h)
+        (h + [0].pack(TEMPLATE_HANDLE)).unpack1(TEMPLATE_HANDLE)
       end
 
+      TEMPLATE_DWORD = 'V'
+
       def packdw(dw)
-        [dw].pack('V')
+        [dw].pack(TEMPLATE_DWORD)
       end
 
       def unpackdw(dw)
-        dw += [0].pack('V')
-        dw.unpack('V')[0]
+        (dw + [0].pack(TEMPLATE_DWORD)).unpack1(TEMPLATE_DWORD)
       end
 
+      TEMPLATE_QWORD = 'Q<'
+
       def packqw(qw)
-        [ qw & 0xFFFFFFFF, qw >> 32 ].pack('VV')
+        [qw].pack(TEMPLATE_QWORD)
       end
 
       def unpackqw(qw)
-        qw = qw.unpack('VV')
-        (qw[1] << 32) | qw[0]
+        (qw + [0].pack(TEMPLATE_QWORD)).unpack1(TEMPLATE_QWORD)
       end
 
       def make_wstr(str)
@@ -317,7 +326,7 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
         size = packdw(0)
         name = make_wstr(name)
         check RegQueryValueExW.call(hkey, name, 0, type, 0, size)
-        data = "\0".force_encoding('ASCII-8BIT') * unpackdw(size)
+        data = "\0".b * unpackdw(size)
         check RegQueryValueExW.call(hkey, name, 0, type, data, size)
         [ unpackdw(type), data[0, unpackdw(size)] ]
       end
@@ -332,11 +341,11 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
       end
 
       def DeleteValue(hkey, name)
-        check RegDeleteValue.call(hkey, make_wstr(name))
+        check RegDeleteValueW.call(hkey, make_wstr(name))
       end
 
       def DeleteKey(hkey, name)
-        check RegDeleteKey.call(hkey, make_wstr(name))
+        check RegDeleteKeyW.call(hkey, make_wstr(name))
       end
 
       def FlushKey(hkey)
@@ -372,8 +381,7 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
     def self.expand_environ(str)
       str.gsub(Regexp.compile("%([^%]+)%".encode(str.encoding))) {
         v = $1.encode(LOCALE)
-        (e = ENV[v] || ENV[v.upcase]; e.encode(str.encoding) if e) ||
-        $&
+        (ENV[v] || ENV[v.upcase])&.encode(str.encoding) || $&
       }
     end
 
@@ -383,7 +391,6 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
       REG_RESOURCE_LIST REG_FULL_RESOURCE_DESCRIPTOR
       REG_RESOURCE_REQUIREMENTS_LIST REG_QWORD
     ].inject([]) do |ary, type|
-      type.freeze
       ary[Constants.const_get(type)] = type
       ary
     end.freeze
@@ -574,9 +581,9 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
         begin
           type, data = read(subkey)
         rescue Error
-          next
+        else
+          yield subkey, type, data
         end
-        yield subkey, type, data
         index += 1
       end
       index
@@ -634,7 +641,7 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
     #    Array of String
     # :REG_DWORD, REG_DWORD_BIG_ENDIAN, REG_QWORD
     #    Integer
-    # :REG_BINARY
+    # :REG_BINARY, REG_NONE
     #    String (contains binary data)
     #
     # When rtype is specified, the value type must be included by
@@ -642,36 +649,38 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
     def read(name, *rtype)
       type, data = API.QueryValue(@hkey, name)
       unless rtype.empty? or rtype.include?(type)
-        raise TypeError, "Type mismatch (expect #{rtype.inspect} but #{type} present)"
+        raise TypeError, "Type mismatch (expect [#{
+          rtype.map{|t|Registry.type2name(t)}.join(', ')}] but #{
+          Registry.type2name(type)} present)"
       end
       case type
       when REG_SZ, REG_EXPAND_SZ
         [ type, data.encode(name.encoding, WCHAR).chop ]
       when REG_MULTI_SZ
         [ type, data.encode(name.encoding, WCHAR).split(/\0/) ]
-      when REG_BINARY
+      when REG_BINARY, REG_NONE
         [ type, data ]
       when REG_DWORD
         [ type, API.unpackdw(data) ]
       when REG_DWORD_BIG_ENDIAN
-        [ type, data.unpack('N')[0] ]
+        [ type, data.unpack1('N') ]
       when REG_QWORD
         [ type, API.unpackqw(data) ]
       else
-        raise TypeError, "Type #{type} is not supported."
+        raise TypeError, "Type #{Registry.type2name(type)} is not supported."
       end
     end
 
     #
     # Read a registry value named name and return its value data.
-    # The class of value is same as #read method returns.
+    # The class of the value is the same as the #read method returns.
     #
     # If the value type is REG_EXPAND_SZ, returns value data whose environment
     # variables are replaced.
     # If the value type is neither REG_SZ, REG_MULTI_SZ, REG_DWORD,
     # REG_DWORD_BIG_ENDIAN, nor REG_QWORD, TypeError is raised.
     #
-    # The meaning of rtype is same as #read method.
+    # The meaning of rtype is the same as for the #read method.
     #
     def [](name, *rtype)
       type, data = read(name, *rtype)
@@ -681,7 +690,7 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
       when REG_EXPAND_SZ
         Registry.expand_environ(data)
       else
-        raise TypeError, "Type #{type} is not supported."
+        raise TypeError, "Type #{Registry.type2name(type)} is not supported."
       end
     end
 
@@ -737,15 +746,12 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
     # method returns.
     #
     def write(name, type, data)
-      termsize = 0
       case type
       when REG_SZ, REG_EXPAND_SZ
-        data = data.encode(WCHAR)
-        termsize = WCHAR_SIZE
+        data = data.encode(WCHAR) << WCHAR_NUL
       when REG_MULTI_SZ
         data = data.to_a.map {|s| s.encode(WCHAR)}.join(WCHAR_NUL) << WCHAR_NUL
-        termsize = WCHAR_SIZE
-      when REG_BINARY
+      when REG_BINARY, REG_NONE
         data = data.to_s
       when REG_DWORD
         data = API.packdw(data.to_i)
@@ -754,9 +760,9 @@ For detail, see the MSDN[http://msdn.microsoft.com/library/en-us/sysinfo/base/pr
       when REG_QWORD
         data = API.packqw(data.to_i)
       else
-        raise TypeError, "Unsupported type #{type}"
+        raise TypeError, "Unsupported type #{Registry.type2name(type)}"
       end
-      API.SetValue(@hkey, name, type, data, data.bytesize + termsize)
+      API.SetValue(@hkey, name, type, data, data.bytesize)
     end
 
     #
