@@ -1,6 +1,7 @@
 require "test/unit"
 
 require "error_highlight"
+require "did_you_mean"
 require "tempfile"
 
 class ErrorHighlightTest < Test::Unit::TestCase
@@ -23,21 +24,48 @@ class ErrorHighlightTest < Test::Unit::TestCase
     end
   end
 
+  begin
+    method_not_exist
+  rescue NameError
+    if $!.message.include?("`")
+      def preprocess(msg)
+        msg
+      end
+    else
+      def preprocess(msg)
+        msg.sub("`", "'")
+      end
+    end
+  end
+
   if Exception.method_defined?(:detailed_message)
     def assert_error_message(klass, expected_msg, &blk)
+      omit unless klass < ErrorHighlight::CoreExt
       err = assert_raise(klass, &blk)
-      assert_equal(expected_msg.chomp, err.detailed_message(highlight: false).sub(/ \((?:NoMethod|Name)Error\)/, ""))
+      assert_equal(preprocess(expected_msg).chomp, err.detailed_message(highlight: false).sub(/ \((?:NoMethod|Name)Error\)/, ""))
     end
   else
     def assert_error_message(klass, expected_msg, &blk)
+      omit unless klass < ErrorHighlight::CoreExt
       err = assert_raise(klass, &blk)
-      assert_equal(expected_msg.chomp, err.message)
+      assert_equal(preprocess(expected_msg).chomp, err.message)
     end
   end
 
+  if begin; 1.time; rescue; $!.message.end_with?("an instance of Integer"); end
+    # new message format
+    NEW_MESSAGE_FORMAT = true
+    NIL_RECV_MESSAGE = "nil"
+    ONE_RECV_MESSAGE = "an instance of Integer"
+  else
+    NEW_MESSAGE_FORMAT = false
+    NIL_RECV_MESSAGE = "nil:NilClass"
+    ONE_RECV_MESSAGE = "1:Integer"
+  end
+
   def test_CALL_noarg_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       nil.foo + 1
          ^^^^
@@ -49,7 +77,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_noarg_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
         .foo + 1
         ^^^^
@@ -62,7 +90,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_noarg_3
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
         foo + 1
         ^^^
@@ -75,7 +103,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_noarg_4
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       (nil).foo + 1
            ^^^^
@@ -87,7 +115,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_arg_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       nil.foo (42)
          ^^^^
@@ -99,7 +127,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_arg_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
         .foo (
         ^^^^
@@ -114,7 +142,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_arg_3
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
         foo (
         ^^^
@@ -129,7 +157,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_arg_4
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       nil.foo(42)
          ^^^^
@@ -141,7 +169,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_arg_5
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
         .foo(
         ^^^^
@@ -156,7 +184,7 @@ undefined method `foo' for nil:NilClass
 
   def test_CALL_arg_6
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
         foo(
         ^^^
@@ -169,9 +197,18 @@ undefined method `foo' for nil:NilClass
     end
   end
 
+  def test_CALL_arg_7
+    assert_error_message(ArgumentError, <<~END) do
+tried to create Proc object without a block (ArgumentError)
+    END
+
+      Proc.new
+    end
+  end
+
   def test_QCALL_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for 1:Integer
+undefined method `foo' for #{ ONE_RECV_MESSAGE }
 
       1&.foo
        ^^^^^
@@ -183,7 +220,7 @@ undefined method `foo' for 1:Integer
 
   def test_QCALL_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for 1:Integer
+undefined method `foo' for #{ ONE_RECV_MESSAGE }
 
       1&.foo(42)
        ^^^^^
@@ -195,7 +232,7 @@ undefined method `foo' for 1:Integer
 
   def test_CALL_aref_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for nil:NilClass
+undefined method `[]' for #{ NIL_RECV_MESSAGE }
 
       nil [ ]
           ^^^
@@ -207,7 +244,7 @@ undefined method `[]' for nil:NilClass
 
   def test_CALL_aref_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for nil:NilClass
+undefined method `[]' for #{ NIL_RECV_MESSAGE }
 
       nil [0]
           ^^^
@@ -219,7 +256,7 @@ undefined method `[]' for nil:NilClass
 
   def test_CALL_aref_3
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for nil:NilClass
+undefined method `[]' for #{ NIL_RECV_MESSAGE }
     END
 
       nil [
@@ -230,8 +267,9 @@ undefined method `[]' for nil:NilClass
 
   def test_CALL_aref_4
     v = Object.new
+    recv = NEW_MESSAGE_FORMAT ? "an instance of Object" : v.inspect
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for #{ v.inspect }
+undefined method `[]' for #{ recv }
 
       v &.[](0)
         ^^^^
@@ -243,7 +281,7 @@ undefined method `[]' for #{ v.inspect }
 
   def test_CALL_aref_5
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for nil:NilClass
+undefined method `[]' for #{ NIL_RECV_MESSAGE }
 
       (nil)[ ]
            ^^^
@@ -255,7 +293,7 @@ undefined method `[]' for nil:NilClass
 
   def test_CALL_aset
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]=' for nil:NilClass
+undefined method `[]=' for #{ NIL_RECV_MESSAGE }
 
       nil.[]=
          ^^^^
@@ -268,7 +306,7 @@ undefined method `[]=' for nil:NilClass
   def test_CALL_op_asgn
     v = nil
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       v += 42
         ^
@@ -280,7 +318,7 @@ undefined method `+' for nil:NilClass
 
   def test_CALL_special_call_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `call' for nil:NilClass
+undefined method `call' for #{ NIL_RECV_MESSAGE }
     END
 
       nil.()
@@ -289,7 +327,7 @@ undefined method `call' for nil:NilClass
 
   def test_CALL_special_call_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `call' for nil:NilClass
+undefined method `call' for #{ NIL_RECV_MESSAGE }
     END
 
       nil.(42)
@@ -298,7 +336,7 @@ undefined method `call' for nil:NilClass
 
   def test_CALL_send
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       nil.send(:foo, 42)
          ^^^^^
@@ -310,7 +348,7 @@ undefined method `foo' for nil:NilClass
 
   def test_ATTRASGN_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]=' for nil:NilClass
+undefined method `[]=' for #{ NIL_RECV_MESSAGE }
 
       nil [ ] = 42
           ^^^^^
@@ -322,7 +360,7 @@ undefined method `[]=' for nil:NilClass
 
   def test_ATTRASGN_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]=' for nil:NilClass
+undefined method `[]=' for #{ NIL_RECV_MESSAGE }
 
       nil [0] = 42
           ^^^^^
@@ -334,7 +372,7 @@ undefined method `[]=' for nil:NilClass
 
   def test_ATTRASGN_3
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo=' for nil:NilClass
+undefined method `foo=' for #{ NIL_RECV_MESSAGE }
 
       nil.foo = 42
          ^^^^^^
@@ -346,7 +384,7 @@ undefined method `foo=' for nil:NilClass
 
   def test_ATTRASGN_4
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]=' for nil:NilClass
+undefined method `[]=' for #{ NIL_RECV_MESSAGE }
 
       (nil)[0] = 42
            ^^^^^
@@ -358,7 +396,7 @@ undefined method `[]=' for nil:NilClass
 
   def test_ATTRASGN_5
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo=' for nil:NilClass
+undefined method `foo=' for #{ NIL_RECV_MESSAGE }
 
       (nil).foo = 42
            ^^^^^^
@@ -370,7 +408,7 @@ undefined method `foo=' for nil:NilClass
 
   def test_OPCALL_binary_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       nil + 42
           ^
@@ -382,7 +420,7 @@ undefined method `+' for nil:NilClass
 
   def test_OPCALL_binary_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       nil + # comment
           ^
@@ -395,7 +433,7 @@ undefined method `+' for nil:NilClass
 
   def test_OPCALL_binary_3
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       (nil) + 42
             ^
@@ -407,7 +445,7 @@ undefined method `+' for nil:NilClass
 
   def test_OPCALL_unary_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+@' for nil:NilClass
+undefined method `+@' for #{ NIL_RECV_MESSAGE }
 
       + nil
       ^
@@ -419,7 +457,7 @@ undefined method `+@' for nil:NilClass
 
   def test_OPCALL_unary_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+@' for nil:NilClass
+undefined method `+@' for #{ NIL_RECV_MESSAGE }
 
       +(nil)
       ^
@@ -431,7 +469,7 @@ undefined method `+@' for nil:NilClass
 
   def test_FCALL_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       nil.instance_eval { foo() }
                           ^^^
@@ -443,7 +481,7 @@ undefined method `foo' for nil:NilClass
 
   def test_FCALL_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       nil.instance_eval { foo(42) }
                           ^^^
@@ -455,7 +493,7 @@ undefined method `foo' for nil:NilClass
 
   def test_VCALL_2
     assert_error_message(NameError, <<~END) do
-undefined local variable or method `foo' for nil:NilClass
+undefined local variable or method `foo' for #{ NIL_RECV_MESSAGE }
 
       nil.instance_eval { foo }
                           ^^^
@@ -469,7 +507,7 @@ undefined local variable or method `foo' for nil:NilClass
     v = nil
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for nil:NilClass
+undefined method `[]' for #{ NIL_RECV_MESSAGE }
 
       v [0] += 42
         ^^^
@@ -483,7 +521,7 @@ undefined method `[]' for nil:NilClass
     v = nil
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for nil:NilClass
+undefined method `[]' for #{ NIL_RECV_MESSAGE }
 
       v [0] += # comment
         ^^^
@@ -498,7 +536,7 @@ undefined method `[]' for nil:NilClass
     v = nil
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for nil:NilClass
+undefined method `[]' for #{ NIL_RECV_MESSAGE }
     END
 
       v [
@@ -512,7 +550,7 @@ undefined method `[]' for nil:NilClass
     v = nil
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `[]' for nil:NilClass
+undefined method `[]' for #{ NIL_RECV_MESSAGE }
 
       (v)[0] += 42
          ^^^
@@ -527,7 +565,7 @@ undefined method `[]' for nil:NilClass
     def v.[](x); nil; end
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       v [0] += 42
             ^
@@ -542,7 +580,7 @@ undefined method `+' for nil:NilClass
     def v.[](x); nil; end
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       v [0 ] += # comment
              ^
@@ -558,7 +596,7 @@ undefined method `+' for nil:NilClass
     def v.[](x); nil; end
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
     END
 
       v [
@@ -573,7 +611,7 @@ undefined method `+' for nil:NilClass
     def v.[](x); nil; end
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       (v)[0] += 42
              ^
@@ -648,7 +686,7 @@ undefined method `[]=' for #{ v.inspect }
     v = nil
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       v.foo += 42
        ^^^^
@@ -662,7 +700,7 @@ undefined method `foo' for nil:NilClass
     v = nil
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       v.foo += # comment
        ^^^^
@@ -677,7 +715,7 @@ undefined method `foo' for nil:NilClass
     v = nil
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
 
       (v).foo += 42
          ^^^^
@@ -692,7 +730,7 @@ undefined method `foo' for nil:NilClass
     def v.foo; nil; end
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       v.foo += 42
             ^
@@ -707,7 +745,7 @@ undefined method `+' for nil:NilClass
     def v.foo; nil; end
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       v.foo += # comment
             ^
@@ -723,7 +761,7 @@ undefined method `+' for nil:NilClass
     def v.foo; nil; end
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       (v).foo += 42
               ^
@@ -816,6 +854,54 @@ uninitialized constant ErrorHighlightTest::NotDefined
     end
   end
 
+  def test_COLON2_3
+    assert_error_message(NameError, <<~END) do
+uninitialized constant ErrorHighlightTest::NotDefined
+
+      ErrorHighlightTest::NotDefined::Foo
+                        ^^^^^^^^^^^^
+    END
+
+      ErrorHighlightTest::NotDefined::Foo
+    end
+  end
+
+  def test_COLON2_4
+    assert_error_message(NameError, <<~END) do
+uninitialized constant ErrorHighlightTest::NotDefined
+
+      ::ErrorHighlightTest::NotDefined::Foo
+                          ^^^^^^^^^^^^
+    END
+
+      ::ErrorHighlightTest::NotDefined::Foo
+    end
+  end
+
+  if ErrorHighlight.const_get(:Spotter).const_get(:OPT_GETCONSTANT_PATH)
+    def test_COLON2_5
+      # Unfortunately, we cannot identify which `NotDefined` caused the NameError
+      assert_error_message(NameError, <<~END) do
+  uninitialized constant ErrorHighlightTest::NotDefined
+      END
+
+        ErrorHighlightTest::NotDefined::NotDefined
+      end
+    end
+  else
+    def test_COLON2_5
+      assert_error_message(NameError, <<~END) do
+uninitialized constant ErrorHighlightTest::NotDefined
+
+        ErrorHighlightTest::NotDefined::NotDefined
+                          ^^^^^^^^^^^^
+      END
+
+        ErrorHighlightTest::NotDefined::NotDefined
+      end
+    end
+  end
+
   def test_COLON3
     assert_error_message(NameError, <<~END) do
 uninitialized constant NotDefined
@@ -869,7 +955,7 @@ uninitialized constant ErrorHighlightTest::OP_CDECL_TEST::NotDefined
 
   def test_OP_CDECL_op_1
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       OP_CDECL_TEST::Nil += 1
                          ^
@@ -881,7 +967,7 @@ undefined method `+' for nil:NilClass
 
   def test_OP_CDECL_op_2
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
       OP_CDECL_TEST::Nil += # comment
                          ^
@@ -894,7 +980,7 @@ undefined method `+' for nil:NilClass
 
   def test_OP_CDECL_op_3
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for nil:NilClass
+undefined method `+' for #{ NIL_RECV_MESSAGE }
 
         Nil += 1
             ^
@@ -918,8 +1004,9 @@ uninitialized constant NotDefined
   end
 
   def test_OP_CDECL_toplevel_2
+    recv = NEW_MESSAGE_FORMAT ? "class ErrorHighlightTest" : "ErrorHighlightTest:Class"
     assert_error_message(NoMethodError, <<~END) do
-undefined method `+' for ErrorHighlightTest:Class
+undefined method `+' for #{ recv }
 
       ::ErrorHighlightTest += 1
                            ^
@@ -980,18 +1067,16 @@ local variable `foo' is not defined for #{ b.inspect }
 
   def test_multibyte
     assert_error_message(NoMethodError, <<~END) do
-undefined method `あいうえお' for nil:NilClass
+undefined method `あいうえお' for #{ NIL_RECV_MESSAGE }
     END
 
       nil.あいうえお
     end
   end
 
-  if false
-
   def test_args_CALL_1
     assert_error_message(TypeError, <<~END) do
-nil can't be coerced into Integer
+nil can't be coerced into Integer (TypeError)
 
       1.+(nil)
           ^^^
@@ -1004,7 +1089,7 @@ nil can't be coerced into Integer
   def test_args_CALL_2
     v = []
     assert_error_message(TypeError, <<~END) do
-no implicit conversion from nil to integer
+no implicit conversion from nil to integer (TypeError)
 
       v[nil]
         ^^^
@@ -1017,7 +1102,7 @@ no implicit conversion from nil to integer
   def test_args_ATTRASGN_1
     v = []
     assert_error_message(ArgumentError, <<~END) do
-wrong number of arguments (given 1, expected 2..3)
+wrong number of arguments (given 1, expected 2..3) (ArgumentError)
 
       v [ ] = 1
          ^^^^^^
@@ -1030,7 +1115,7 @@ wrong number of arguments (given 1, expected 2..3)
   def test_args_ATTRASGN_2
     v = []
     assert_error_message(TypeError, <<~END) do
-no implicit conversion from nil to integer
+no implicit conversion from nil to integer (TypeError)
 
       v [nil] = 1
          ^^^^^^^^
@@ -1042,7 +1127,7 @@ no implicit conversion from nil to integer
 
   def test_args_ATTRASGN_3
     assert_error_message(TypeError, <<~END) do
-no implicit conversion of String into Integer
+no implicit conversion of String into Integer (TypeError)
 
       $stdin.lineno = "str"
                       ^^^^^
@@ -1054,7 +1139,7 @@ no implicit conversion of String into Integer
 
   def test_args_OPCALL
     assert_error_message(TypeError, <<~END) do
-nil can't be coerced into Integer
+nil can't be coerced into Integer (TypeError)
 
       1 + nil
           ^^^
@@ -1066,7 +1151,7 @@ nil can't be coerced into Integer
 
   def test_args_FCALL_1
     assert_error_message(TypeError, <<~END) do
-no implicit conversion of Symbol into String
+no implicit conversion of Symbol into String (TypeError)
 
       "str".instance_eval { gsub("foo", :sym) }
                                  ^^^^^^^^^^^
@@ -1078,7 +1163,7 @@ no implicit conversion of Symbol into String
 
   def test_args_FCALL_2
     assert_error_message(TypeError, <<~END) do
-no implicit conversion of Symbol into String
+no implicit conversion of Symbol into String (TypeError)
 
       "str".instance_eval { gsub "foo", :sym }
                                  ^^^^^^^^^^^
@@ -1092,7 +1177,7 @@ no implicit conversion of Symbol into String
     v = []
 
     assert_error_message(TypeError, <<~END) do
-no implicit conversion from nil to integer
+no implicit conversion from nil to integer (TypeError)
 
       v [nil] += 42
          ^^^^^^^^^^
@@ -1106,7 +1191,7 @@ no implicit conversion from nil to integer
     v = []
 
     assert_error_message(ArgumentError, <<~END) do
-wrong number of arguments (given 0, expected 1..2)
+wrong number of arguments (given 0, expected 1..2) (ArgumentError)
 
       v [ ] += 42
          ^^^^^^^^
@@ -1120,7 +1205,7 @@ wrong number of arguments (given 0, expected 1..2)
     v = [1]
 
     assert_error_message(TypeError, <<~END) do
-nil can't be coerced into Integer
+nil can't be coerced into Integer (TypeError)
 
       v [0] += nil
          ^^^^^^^^^
@@ -1135,7 +1220,7 @@ nil can't be coerced into Integer
     def v.foo; 1; end
 
     assert_error_message(TypeError, <<~END) do
-nil can't be coerced into Integer
+nil can't be coerced into Integer (TypeError)
 
       v.foo += nil
                ^^^
@@ -1145,8 +1230,6 @@ nil can't be coerced into Integer
     end
   end
 
-  end
-
   def test_custom_formatter
     custom_formatter = Object.new
     def custom_formatter.message_for(spot)
@@ -1156,9 +1239,9 @@ nil can't be coerced into Integer
     original_formatter, ErrorHighlight.formatter = ErrorHighlight.formatter, custom_formatter
 
     assert_error_message(NoMethodError, <<~END) do
-undefined method `time' for 1:Integer
+undefined method `time' for #{ ONE_RECV_MESSAGE }
 
-{:first_lineno=>#{ __LINE__ + 3 }, :first_column=>7, :last_lineno=>#{ __LINE__ + 3 }, :last_column=>12, :snippet=>"      1.time {}\\n"}
+#{{ first_lineno: __LINE__ + 3, first_column: 7, last_lineno: __LINE__ + 3, last_column: 12, snippet: "      1.time {}\n" }.inspect}
     END
 
       1.time {}
@@ -1174,7 +1257,7 @@ undefined method `time' for 1:Integer
       tmp.close
 
       assert_error_message(NoMethodError, <<~END.gsub("_", "\t")) do
-undefined method `time' for 1:Integer
+undefined method `time' for #{ ONE_RECV_MESSAGE }
 
 _ _1.time {}
 _ _ ^^^^^
@@ -1191,7 +1274,7 @@ _ _ ^^^^^
       tmp.close
 
       assert_error_message(NoMethodError, <<~END) do
-undefined method `time' for 1:Integer
+undefined method `time' for #{ ONE_RECV_MESSAGE }
 
 1.time {}
  ^^^^^
@@ -1204,7 +1287,7 @@ undefined method `time' for 1:Integer
 
   def test_simulate_funcallv_from_embedded_ruby
     assert_error_message(NoMethodError, <<~END) do
-undefined method `foo' for nil:NilClass
+undefined method `foo' for #{ NIL_RECV_MESSAGE }
     END
 
       nil.foo + 1
@@ -1219,8 +1302,9 @@ undefined method `foo' for nil:NilClass
       tmp << "module Dummy\nend\n"
       tmp.close
 
+      recv = NEW_MESSAGE_FORMAT ? "an instance of String" : '"dummy":String'
       assert_error_message(NameError, <<~END) do
-        undefined local variable or method `foo' for "dummy":String
+        undefined local variable or method `foo' for #{ recv }
       END
 
         "dummy".instance_eval do
@@ -1257,4 +1341,41 @@ undefined method `foo' for nil:NilClass
     assert_equal(22, spot[:last_column])
     assert_equal("      raise_name_error\n", spot[:snippet])
   end
+
+  def test_spot_with_node
+    omit unless RubyVM::AbstractSyntaxTree.respond_to?(:node_id_for_backtrace_location)
+
+    # We can't revisit instruction sequences to find node ids if the prism
+    # compiler was used instead of the parse.y compiler. In that case, we'll
+    # omit some tests.
+    omit if RubyVM::InstructionSequence.compile("").to_a[4][:parser] == :prism
+
+    begin
+      raise_name_error
+    rescue NameError => exc
+    end
+
+    bl = exc.backtrace_locations.first
+    expected_spot = ErrorHighlight.spot(exc, backtrace_location: bl)
+    ast = RubyVM::AbstractSyntaxTree.parse_file(__FILE__, keep_script_lines: true)
+    node_id = RubyVM::AbstractSyntaxTree.node_id_for_backtrace_location(bl)
+    node = find_node_by_id(ast, node_id)
+    actual_spot = ErrorHighlight.spot(node)
+
+    assert_equal expected_spot, actual_spot
+  end
+
+  private
+
+  def find_node_by_id(node, node_id)
+    return node if node.node_id == node_id
+
+    node.children.each do |child|
+      next unless child.is_a?(RubyVM::AbstractSyntaxTree::Node)
+      found = find_node_by_id(child, node_id)
+      return found if found
+    end
+
+    return false
+  end
 end
