@@ -40,16 +40,93 @@ class TestGc < Test::Unit::TestCase
   end
 
   def test_enable_disable
-    GC.enable
-    assert_equal(false, GC.enable)
-    assert_equal(false, GC.disable)
-    assert_equal(true, GC.disable)
-    assert_equal(true, GC.disable)
-    assert_nil(GC.start)
-    assert_equal(true, GC.enable)
-    assert_equal(false, GC.enable)
+    EnvUtil.without_gc do
+      GC.enable
+      assert_equal(false, GC.enable)
+      assert_equal(false, GC.disable)
+      assert_equal(true, GC.disable)
+      assert_equal(true, GC.disable)
+      assert_nil(GC.start)
+      assert_equal(true, GC.enable)
+      assert_equal(false, GC.enable)
+    end
+  end
+
+  def test_gc_config_full_mark_by_default
+    omit "unsupoported platform/GC" unless defined?(GC.config)
+
+    config = GC.config
+    assert_not_empty(config)
+    assert_true(config[:rgengc_allow_full_mark])
+  end
+
+  def test_gc_config_invalid_args
+    omit "unsupoported platform/GC" unless defined?(GC.config)
+
+    assert_raise(ArgumentError) { GC.config(0) }
+  end
+
+  def test_gc_config_setting_returns_updated_config_hash
+    omit "unsupoported platform/GC" unless defined?(GC.config)
+
+    old_value = GC.config[:rgengc_allow_full_mark]
+    assert_true(old_value)
+
+    new_value = GC.config(rgengc_allow_full_mark: false)[:rgengc_allow_full_mark]
+    assert_false(new_value)
+    new_value = GC.config(rgengc_allow_full_mark: nil)[:rgengc_allow_full_mark]
+    assert_false(new_value)
   ensure
+    GC.config(rgengc_allow_full_mark: old_value)
+    GC.start
+  end
+
+  def test_gc_config_setting_returns_nil_for_missing_keys
+    omit "unsupoported platform/GC" unless defined?(GC.config)
+
+    missing_value = GC.config(no_such_key: true)[:no_such_key]
+    assert_nil(missing_value)
+  ensure
+    GC.config(full_mark: true)
+    GC.start
+  end
+
+  def test_gc_config_disable_major
+    omit "unsupoported platform/GC" unless defined?(GC.config)
+
     GC.enable
+    GC.start
+
+    GC.config(rgengc_allow_full_mark: false)
+    major_count = GC.stat[:major_gc_count]
+    minor_count = GC.stat[:minor_gc_count]
+
+    arr = []
+    (GC.stat_heap[0][:heap_eden_slots] * 2).times do
+      arr << Object.new
+      Object.new
+    end
+
+    assert_equal(major_count, GC.stat[:major_gc_count])
+    assert_operator(minor_count, :<=, GC.stat[:minor_gc_count])
+    assert_nil(GC.start)
+  ensure
+    GC.config(rgengc_allow_full_mark: true)
+    GC.start
+  end
+
+  def test_gc_config_disable_major_gc_start_always_works
+    omit "unsupoported platform/GC" unless defined?(GC.config)
+
+    GC.config(full_mark: false)
+
+    major_count = GC.stat[:major_gc_count]
+    GC.start
+
+    assert_operator(major_count, :<, GC.stat[:major_gc_count])
+  ensure
+    GC.config(full_mark: true)
+    GC.start
   end
 
   def test_start_full_mark
@@ -131,10 +208,9 @@ class TestGc < Test::Unit::TestCase
     # marking_time + sweeping_time could differ from time by 1 because they're stored in nanoseconds
     assert_in_delta stat[:time], stat[:marking_time] + stat[:sweeping_time], 1
     assert_equal stat[:total_allocated_pages], stat[:heap_allocated_pages] + stat[:total_freed_pages]
-    assert_operator stat[:heap_sorted_length], :>=, stat[:heap_eden_pages] + stat[:heap_allocatable_pages], "stat is: " + stat.inspect
     assert_equal stat[:heap_available_slots], stat[:heap_live_slots] + stat[:heap_free_slots] + stat[:heap_final_slots]
     assert_equal stat[:heap_live_slots], stat[:total_allocated_objects] - stat[:total_freed_objects] - stat[:heap_final_slots]
-    assert_equal stat[:heap_allocated_pages], stat[:heap_eden_pages] + stat[:heap_tomb_pages]
+    assert_equal stat[:heap_allocated_pages], stat[:heap_eden_pages]
 
     if use_rgengc?
       assert_equal stat[:count], stat[:major_gc_count] + stat[:minor_gc_count]
@@ -150,18 +226,16 @@ class TestGc < Test::Unit::TestCase
     GC.stat_heap(0, stat_heap)
     GC.stat(stat)
 
-    GC::INTERNAL_CONSTANTS[:SIZE_POOL_COUNT].times do |i|
-      GC.stat_heap(i, stat_heap)
-      GC.stat(stat)
+    GC::INTERNAL_CONSTANTS[:HEAP_COUNT].times do |i|
+      EnvUtil.without_gc do
+        GC.stat_heap(i, stat_heap)
+        GC.stat(stat)
+      end
 
-      assert_equal GC::INTERNAL_CONSTANTS[:RVALUE_SIZE] * (2**i), stat_heap[:slot_size]
-      assert_operator stat_heap[:heap_allocatable_pages], :<=, stat[:heap_allocatable_pages]
+      assert_equal (GC::INTERNAL_CONSTANTS[:BASE_SLOT_SIZE] + GC::INTERNAL_CONSTANTS[:RVALUE_OVERHEAD]) * (2**i), stat_heap[:slot_size]
       assert_operator stat_heap[:heap_eden_pages], :<=, stat[:heap_eden_pages]
       assert_operator stat_heap[:heap_eden_slots], :>=, 0
-      assert_operator stat_heap[:heap_tomb_pages], :<=, stat[:heap_tomb_pages]
-      assert_operator stat_heap[:heap_tomb_slots], :>=, 0
       assert_operator stat_heap[:total_allocated_pages], :>=, 0
-      assert_operator stat_heap[:total_freed_pages], :>=, 0
       assert_operator stat_heap[:force_major_gc_count], :>=, 0
       assert_operator stat_heap[:force_incremental_marking_finish_count], :>=, 0
       assert_operator stat_heap[:total_allocated_objects], :>=, 0
@@ -174,19 +248,19 @@ class TestGc < Test::Unit::TestCase
     assert_equal stat_heap[:slot_size], GC.stat_heap(0)[:slot_size]
 
     assert_raise(ArgumentError) { GC.stat_heap(-1) }
-    assert_raise(ArgumentError) { GC.stat_heap(GC::INTERNAL_CONSTANTS[:SIZE_POOL_COUNT]) }
+    assert_raise(ArgumentError) { GC.stat_heap(GC::INTERNAL_CONSTANTS[:HEAP_COUNT]) }
   end
 
   def test_stat_heap_all
+    omit "flaky with RJIT, which allocates objects itself" if defined?(RubyVM::RJIT) && RubyVM::RJIT.enabled?
     stat_heap_all = {}
     stat_heap = {}
+    # Initialize to prevent GC in future calls
+    GC.stat_heap(0, stat_heap)
+    GC.stat_heap(nil, stat_heap_all)
 
-    2.times do
-      GC.stat_heap(0, stat_heap)
+    GC::INTERNAL_CONSTANTS[:HEAP_COUNT].times do |i|
       GC.stat_heap(nil, stat_heap_all)
-    end
-
-    GC::INTERNAL_CONSTANTS[:SIZE_POOL_COUNT].times do |i|
       GC.stat_heap(i, stat_heap)
 
       # Remove keys that can vary between invocations
@@ -215,22 +289,35 @@ class TestGc < Test::Unit::TestCase
       hash.each { |k, v| stat_heap_sum[k] += v }
     end
 
-    assert_equal stat[:heap_allocatable_pages], stat_heap_sum[:heap_allocatable_pages]
     assert_equal stat[:heap_eden_pages], stat_heap_sum[:heap_eden_pages]
-    assert_equal stat[:heap_tomb_pages], stat_heap_sum[:heap_tomb_pages]
-    assert_equal stat[:heap_available_slots], stat_heap_sum[:heap_eden_slots] + stat_heap_sum[:heap_tomb_slots]
-    assert_equal stat[:total_allocated_pages], stat_heap_sum[:total_allocated_pages]
-    assert_equal stat[:total_freed_pages], stat_heap_sum[:total_freed_pages]
+    assert_equal stat[:heap_available_slots], stat_heap_sum[:heap_eden_slots]
     assert_equal stat[:total_allocated_objects], stat_heap_sum[:total_allocated_objects]
     assert_equal stat[:total_freed_objects], stat_heap_sum[:total_freed_objects]
   end
 
+  def test_measure_total_time
+    assert_separately([], __FILE__, __LINE__, <<~RUBY)
+      GC.measure_total_time = false
+
+      time_before = GC.stat(:time)
+
+      # Generate some garbage
+      Random.new.bytes(100 * 1024 * 1024)
+      GC.start
+
+      time_after = GC.stat(:time)
+
+      # If time measurement is disabled, the time stat should not change
+      assert_equal time_before, time_after
+    RUBY
+  end
+
   def test_latest_gc_info
     omit 'stress' if GC.stress
 
     assert_separately([], __FILE__, __LINE__, <<-'RUBY')
       GC.start
-      count = GC.stat(:heap_free_slots) + GC.stat(:heap_allocatable_pages) * GC::INTERNAL_CONSTANTS[:HEAP_PAGE_OBJ_LIMIT]
+      count = GC.stat(:heap_free_slots) + GC.stat(:heap_allocatable_slots)
       count.times{ "a" + "b" }
       assert_equal :newobj, GC.latest_gc_info[:gc_by]
     RUBY
@@ -275,19 +362,22 @@ class TestGc < Test::Unit::TestCase
       objects.append(100.times.map { '*' })
     end
 
-    # We need to ensure that no GC gets ran before the call to GC.start since
-    # it would trigger a major GC. Assertions could allocate objects and
-    # trigger a GC so we don't run assertions until we perform the major GC.
-    need_major_by = GC.latest_gc_info(:need_major_by)
-    GC.start(full_mark: false) # should be upgraded to major
-    major_by = GC.latest_gc_info(:major_by)
+    EnvUtil.without_gc do
+      # We need to ensure that no GC gets ran before the call to GC.start since
+      # it would trigger a major GC. Assertions could allocate objects and
+      # trigger a GC so we don't run assertions until we perform the major GC.
+      need_major_by = GC.latest_gc_info(:need_major_by)
+      GC.start(full_mark: false) # should be upgraded to major
+      major_by = GC.latest_gc_info(:major_by)
 
-    assert_not_nil(need_major_by)
-    assert_not_nil(major_by)
+      assert_not_nil(need_major_by)
+      assert_not_nil(major_by)
+    end
   end
 
   def test_latest_gc_info_weak_references_count
     assert_separately([], __FILE__, __LINE__, <<~RUBY)
+      GC.disable
       count = 10_000
       # Some weak references may be created, so allow some margin of error
       error_tolerance = 100
@@ -327,7 +417,8 @@ class TestGc < Test::Unit::TestCase
       # Run full GC again to collect stats about weak references
       GC.start
 
-      assert_equal(0, wmap.size)
+      # Sometimes the WeakMap has one element, which might be held on by registers.
+      assert_operator(wmap.size, :<=, 1)
 
       assert_operator(GC.latest_gc_info(:weak_references_count), :<=, before_weak_references_count - count + error_tolerance)
       assert_operator(GC.latest_gc_info(:retained_weak_references_count), :<=, before_retained_weak_references_count - count + error_tolerance)
@@ -441,114 +532,40 @@ class TestGc < Test::Unit::TestCase
   end
 
   def test_gc_parameter_init_slots
-    assert_separately([], __FILE__, __LINE__, <<~RUBY)
+    assert_separately([], __FILE__, __LINE__, <<~RUBY, timeout: 60)
       # Constant from gc.c.
       GC_HEAP_INIT_SLOTS = 10_000
-      GC.stat_heap.each do |_, s|
-        multiple = s[:slot_size] / (GC::INTERNAL_CONSTANTS[:BASE_SLOT_SIZE] + GC::INTERNAL_CONSTANTS[:RVALUE_OVERHEAD])
-        # Allocatable pages are assumed to have lost 1 slot due to alignment.
-        slots_per_page = (GC::INTERNAL_CONSTANTS[:HEAP_PAGE_OBJ_LIMIT] / multiple) - 1
 
-        total_slots = s[:heap_eden_slots] + s[:heap_allocatable_pages] * slots_per_page
-        assert_operator(total_slots, :>=, GC_HEAP_INIT_SLOTS, s)
+      gc_count = GC.stat(:count)
+      # Fill up all of the size pools to the init slots
+      GC::INTERNAL_CONSTANTS[:HEAP_COUNT].times do |i|
+        capa = (GC.stat_heap(i, :slot_size) - GC::INTERNAL_CONSTANTS[:RVALUE_OVERHEAD] - (2 * RbConfig::SIZEOF["void*"])) / RbConfig::SIZEOF["void*"]
+        while GC.stat_heap(i, :heap_eden_slots) < GC_HEAP_INIT_SLOTS
+          Array.new(capa)
+        end
       end
+
+      assert_equal gc_count, GC.stat(:count)
     RUBY
 
     env = {}
-    # Make the heap big enough to ensure the heap never needs to grow.
-    sizes = GC.stat_heap.keys.reverse.map { |i| (i + 1) * 100_000 }
+    sizes = GC.stat_heap.keys.reverse.map { 20_000 }
     GC.stat_heap.keys.each do |heap|
       env["RUBY_GC_HEAP_#{heap}_INIT_SLOTS"] = sizes[heap].to_s
     end
-    assert_separately([env, "-W0"], __FILE__, __LINE__, <<~RUBY)
-      SIZES = #{sizes}
-      GC.stat_heap.each do |i, s|
-        multiple = s[:slot_size] / (GC::INTERNAL_CONSTANTS[:BASE_SLOT_SIZE] + GC::INTERNAL_CONSTANTS[:RVALUE_OVERHEAD])
-        # Allocatable pages are assumed to have lost 1 slot due to alignment.
-        slots_per_page = (GC::INTERNAL_CONSTANTS[:HEAP_PAGE_OBJ_LIMIT] / multiple) - 1
-
-        total_slots = s[:heap_eden_slots] + s[:heap_allocatable_pages] * slots_per_page
-
-        # The delta is calculated as follows:
-        #  - For allocated pages, each page can vary by 1 slot due to alignment.
-        #  - For allocatable pages, we can end up with at most 1 extra page of slots.
-        assert_in_delta(SIZES[i], total_slots, s[:heap_eden_pages] + slots_per_page, s)
-      end
-    RUBY
-
-    # Check that the configured sizes are "remembered" across GC invocations.
-    assert_separately([env, "-W0"], __FILE__, __LINE__, <<~RUBY)
+    assert_separately([env, "-W0"], __FILE__, __LINE__, <<~RUBY, timeout: 60)
       SIZES = #{sizes}
 
-      # Fill size pool 0 with transient objects.
-      ary = []
-      while GC.stat_heap(0, :heap_allocatable_pages) != 0
-        ary << Object.new
-      end
-      ary.clear
-      ary = nil
-
-      # Clear all the objects that were allocated.
-      GC.start
-
-      # Check that we still have the same number of slots as initially configured.
-      GC.stat_heap.each do |i, s|
-        multiple = s[:slot_size] / (GC::INTERNAL_CONSTANTS[:BASE_SLOT_SIZE] + GC::INTERNAL_CONSTANTS[:RVALUE_OVERHEAD])
-        # Allocatable pages are assumed to have lost 1 slot due to alignment.
-        slots_per_page = (GC::INTERNAL_CONSTANTS[:HEAP_PAGE_OBJ_LIMIT] / multiple) - 1
-
-        total_slots = s[:heap_eden_slots] + s[:heap_allocatable_pages] * slots_per_page
-
-        # The delta is calculated as follows:
-        #  - For allocated pages, each page can vary by 1 slot due to alignment.
-        #  - For allocatable pages, we can end up with at most 1 extra page of slots.
-        assert_in_delta(SIZES[i], total_slots, s[:heap_eden_pages] + slots_per_page, s)
-      end
-    RUBY
-
-    # Check that we don't grow the heap in minor GC if we have alloctable pages.
-    env["RUBY_GC_HEAP_FREE_SLOTS_MIN_RATIO"] = "0.3"
-    env["RUBY_GC_HEAP_FREE_SLOTS_GOAL_RATIO"] = "0.99"
-    env["RUBY_GC_HEAP_FREE_SLOTS_MAX_RATIO"] = "1.0"
-    env["RUBY_GC_HEAP_OLDOBJECT_LIMIT_FACTOR"] = "100" # Large value to disable major GC
-    assert_separately([env, "-W0"], __FILE__, __LINE__, <<~RUBY)
-      SIZES = #{sizes}
-
-      # Run a major GC to clear out dead objects.
-      GC.start
-
-      # Disable GC so we can control when GC is ran.
-      GC.disable
-
-      # Run minor GC enough times so that we don't grow the heap because we
-      # haven't yet ran RVALUE_OLD_AGE minor GC cycles.
-      GC::INTERNAL_CONSTANTS[:RVALUE_OLD_AGE].times { GC.start(full_mark: false) }
-
-      # Fill size pool 0 to over 50% full so that the number of allocatable
-      # pages that will be created will be over the number in heap_allocatable_pages
-      # (calculated using RUBY_GC_HEAP_FREE_SLOTS_MIN_RATIO).
-      # 70% was chosen here to guarantee that.
-      ary = []
-      while GC.stat_heap(0, :heap_allocatable_pages) >
-          (GC.stat_heap(0, :heap_allocatable_pages) + GC.stat_heap(0, :heap_eden_pages)) * 0.3
-        ary << Object.new
+      gc_count = GC.stat(:count)
+      # Fill up all of the size pools to the init slots
+      GC::INTERNAL_CONSTANTS[:HEAP_COUNT].times do |i|
+        capa = (GC.stat_heap(i, :slot_size) - GC::INTERNAL_CONSTANTS[:RVALUE_OVERHEAD] - (2 * RbConfig::SIZEOF["void*"])) / RbConfig::SIZEOF["void*"]
+        while GC.stat_heap(i, :heap_eden_slots) < SIZES[i]
+          Array.new(capa)
+        end
       end
 
-      GC.start(full_mark: false)
-
-      # Check that we still have the same number of slots as initially configured.
-      GC.stat_heap.each do |i, s|
-        multiple = s[:slot_size] / (GC::INTERNAL_CONSTANTS[:BASE_SLOT_SIZE] + GC::INTERNAL_CONSTANTS[:RVALUE_OVERHEAD])
-        # Allocatable pages are assumed to have lost 1 slot due to alignment.
-        slots_per_page = (GC::INTERNAL_CONSTANTS[:HEAP_PAGE_OBJ_LIMIT] / multiple) - 1
-
-        total_slots = s[:heap_eden_slots] + s[:heap_allocatable_pages] * slots_per_page
-
-        # The delta is calculated as follows:
-        #  - For allocated pages, each page can vary by 1 slot due to alignment.
-        #  - For allocatable pages, we can end up with at most 1 extra page of slots.
-        assert_in_delta(SIZES[i], total_slots, s[:heap_eden_pages] + slots_per_page, s)
-      end
+      assert_equal gc_count, GC.stat(:count)
     RUBY
   end
 
@@ -578,6 +595,14 @@ class TestGc < Test::Unit::TestCase
     RUBY
   end
 
+  def test_profiler_raw_data
+    GC::Profiler.enable
+    GC.start
+    assert GC::Profiler.raw_data
+  ensure
+    GC::Profiler.disable
+  end
+
   def test_profiler_total_time
     GC::Profiler.enable
     GC::Profiler.clear
@@ -616,7 +641,7 @@ class TestGc < Test::Unit::TestCase
   def test_thrashing_for_young_objects
     # This test prevents bugs like [Bug #18929]
 
-    assert_separately([], __FILE__, __LINE__, <<-'RUBY')
+    assert_separately([], __FILE__, __LINE__, <<-'RUBY', timeout: 60)
       # Grow the heap
       @ary = 100_000.times.map { Object.new }
 
@@ -640,7 +665,7 @@ class TestGc < Test::Unit::TestCase
 
       # Should not be thrashing in page creation
       assert_equal before_stats[:heap_allocated_pages], after_stats[:heap_allocated_pages], debug_msg
-      assert_equal 0, after_stats[:heap_tomb_pages], debug_msg
+      assert_equal 0, after_stats[:heap_empty_pages], debug_msg
       assert_equal 0, after_stats[:total_freed_pages], debug_msg
       # Only young objects, so should not trigger major GC
       assert_equal before_stats[:major_gc_count], after_stats[:major_gc_count], debug_msg
@@ -649,7 +674,7 @@ class TestGc < Test::Unit::TestCase
 
   def test_gc_internals
     assert_not_nil GC::INTERNAL_CONSTANTS[:HEAP_PAGE_OBJ_LIMIT]
-    assert_not_nil GC::INTERNAL_CONSTANTS[:RVALUE_SIZE]
+    assert_not_nil GC::INTERNAL_CONSTANTS[:BASE_SLOT_SIZE]
   end
 
   def test_sweep_in_finalizer
@@ -740,13 +765,17 @@ class TestGc < Test::Unit::TestCase
   end
 
   def test_gc_disabled_start
-    begin
-      disabled = GC.disable
+    EnvUtil.without_gc do
       c = GC.count
       GC.start
       assert_equal 1, GC.count - c
-    ensure
-      GC.enable unless disabled
+    end
+
+    EnvUtil.without_gc do
+      c = GC.count
+      GC.start(immediate_mark: false, immediate_sweep: false)
+      10_000.times { Object.new }
+      assert_equal 1, GC.count - c
     end
   end
 
@@ -795,6 +824,15 @@ class TestGc < Test::Unit::TestCase
         obj = nil
       end
     end;
+
+    assert_normal_exit "#{<<~"begin;"}\n#{<<~'end;'}", '[Bug #20042]'
+    begin;
+      def (f = Object.new).call = nil # missing ID
+      o = Object.new
+      ObjectSpace.define_finalizer(o, f)
+      o = nil
+      GC.start
+    end;
   end
 
   def test_object_ids_never_repeat
@@ -812,28 +850,26 @@ class TestGc < Test::Unit::TestCase
   end
 
   def test_old_to_young_reference
-    original_gc_disabled = GC.disable
+    EnvUtil.without_gc do
+      require "objspace"
 
-    require "objspace"
+      old_obj = Object.new
+      4.times { GC.start }
 
-    old_obj = Object.new
-    4.times { GC.start }
+      assert_include ObjectSpace.dump(old_obj), '"old":true'
 
-    assert_include ObjectSpace.dump(old_obj), '"old":true'
+      young_obj = Object.new
+      old_obj.instance_variable_set(:@test, young_obj)
 
-    young_obj = Object.new
-    old_obj.instance_variable_set(:@test, young_obj)
+      # Not immediately promoted to old generation
+      3.times do
+        assert_not_include ObjectSpace.dump(young_obj), '"old":true'
+        GC.start
+      end
 
-    # Not immediately promoted to old generation
-    3.times do
-      assert_not_include ObjectSpace.dump(young_obj), '"old":true'
+      # Takes 4 GC to promote to old generation
       GC.start
+      assert_include ObjectSpace.dump(young_obj), '"old":true'
     end
-
-    # Takes 4 GC to promote to old generation
-    GC.start
-    assert_include ObjectSpace.dump(young_obj), '"old":true'
-  ensure
-    GC.enable if !original_gc_disabled
   end
 end
